@@ -8,7 +8,7 @@ package My::RoleA {
 	use Types::Standard -types;
 	
 	multimethod foo => (
-		signature  => [ HashRef ],
+		positional => [ HashRef ],
 		code       => sub { return "A" },
 		alias      => "foo_a",
 	);
@@ -20,7 +20,7 @@ package My::RoleB {
 	use Types::Standard -types;
 	
 	multimethod foo => (
-		signature  => [ ArrayRef ],
+		positional => [ ArrayRef ],
 		code       => sub { return "B" },
 	);
 }
@@ -33,7 +33,7 @@ package My::Class {
 	with qw( My::RoleA My::RoleB );
 	
 	multimethod foo => (
-		signature  => [ HashRef ],
+		positional => [ HashRef ],
 		code       => sub { return "C" },
 	);
 }
